# -*- coding: utf-8 -*-
import inspect
from abjad.tools import abctools
from abjad.tools import expressiontools
from abjad.tools import markuptools
from abjad.tools import mathtools
from abjad.tools import pitchtools
from abjad.tools import schemetools
from abjad.tools import scoretools
from abjad.tools.topleveltools import attach
from abjad.tools.topleveltools import detach
from abjad.tools.topleveltools import inspect_
from abjad.tools.topleveltools import override
from abjad.tools.topleveltools import iterate
from abjad.tools.topleveltools import new


class LabelAgent(abctools.AbjadObject):
    r'''Label agent.

    ..  container:: example

        Labels pitch names:

        ..  container:: example

            ::

                >>> staff = Staff("c'4 e'4 d'4 f'4")
                >>> label(staff).with_pitches()

            ::

                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new Staff {
                    c'4
                        ^ \markup {
                            \small
                                c'
                            }
                    e'4
                        ^ \markup {
                            \small
                                e'
                            }
                    d'4
                        ^ \markup {
                            \small
                                d'
                            }
                    f'4
                        ^ \markup {
                            \small
                                f'
                            }
                }

        ..  container:: example expression

            ::

                >>> staff = Staff("c'4 e'4 d'4 f'4")
                >>> expression = label().with_pitches()
                >>> f(expression)
                expressiontools.Expression(
                    callbacks=[
                        expressiontools.Expression(
                            evaluation_template='abjad.agenttools.LabelAgent',
                            is_initializer=True,
                            ),
                        expressiontools.Expression(
                            evaluation_template='{}.with_pitches()',
                            ),
                        ],
                    proxy_class=agenttools.LabelAgent,
                    )

            ::

                >>> expression(staff)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new Staff {
                    c'4
                        ^ \markup {
                            \small
                                c'
                            }
                    e'4
                        ^ \markup {
                            \small
                                e'
                            }
                    d'4
                        ^ \markup {
                            \small
                                d'
                            }
                    f'4
                        ^ \markup {
                            \small
                                f'
                            }
                }

    ..  container:: example

        Labels durations:

        ..  container:: example

            ::

                >>> staff = Staff("c'4 e'4 d'4 f'4")
                >>> label(staff).with_durations()

            ::

                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new Staff {
                    c'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    e'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    d'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    f'4
                        ^ \markup {
                            \small
                                1/4
                            }
                }

        ..  container:: example expression

            ::

                >>> staff = Staff("c'4 e'4 d'4 f'4")
                >>> expression = label().with_durations()
                >>> f(expression)
                expressiontools.Expression(
                    callbacks=[
                        expressiontools.Expression(
                            evaluation_template='abjad.agenttools.LabelAgent',
                            is_initializer=True,
                            ),
                        expressiontools.Expression(
                            evaluation_template='{}.with_durations()',
                            ),
                        ],
                    proxy_class=agenttools.LabelAgent,
                    )

            ::

                >>> expression(staff)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new Staff {
                    c'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    e'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    d'4
                        ^ \markup {
                            \small
                                1/4
                            }
                    f'4
                        ^ \markup {
                            \small
                                1/4
                            }
                }

    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        '_client',
        '_expression',
        )

    _pc_number_to_color = {
        0: schemetools.SchemeColor('red'),
        1: schemetools.SchemeColor('MediumBlue'),
        2: schemetools.SchemeColor('orange'),
        3: schemetools.SchemeColor('LightSlateBlue'),
        4: schemetools.SchemeColor('ForestGreen'),
        5: schemetools.SchemeColor('MediumOrchid'),
        6: schemetools.SchemeColor('firebrick'),
        7: schemetools.SchemeColor('DeepPink'),
        8: schemetools.SchemeColor('DarkOrange'),
        9: schemetools.SchemeColor('IndianRed'),
        10: schemetools.SchemeColor('CadetBlue'),
        11: schemetools.SchemeColor('SeaGreen'),
        12: schemetools.SchemeColor('LimeGreen'),
        }

    ### INITIALIZER ###

    def __init__(self, client=None):
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        from abjad.tools import spannertools
        prototype = (
            scoretools.Component,
            selectiontools.Selection,
            spannertools.Spanner,
            list,
            type(None),
            )
        if not isinstance(client, prototype):
            message = 'must be component, selection, spanner, list or none:'
            message += ' {!r}.'
            message = message.format(client)
            raise TypeError(message)
        self._client = client
        self._expression = None

    ### PRIVATE METHODS ###

    def _color_leaf(self, leaf, color):
        if isinstance(leaf, (scoretools.Note, scoretools.Chord)):
            override(leaf).accidental.color = color
            override(leaf).beam.color = color
            override(leaf).dots.color = color
            override(leaf).note_head.color = color
            override(leaf).stem.color = color
        elif isinstance(leaf, scoretools.Rest):
            override(leaf).dots.color = color
            override(leaf).rest.color = color
        return leaf

    def _update_expression(self, frame):
        import abjad
        callback = abjad.Expression._frame_to_callback(frame)
        return self._expression.append_callback(callback)

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        r'''Gets client of label agent.

        Returns component, selection, spanner or none.
        '''
        return self._client

    ### PUBLIC METHODS ###

    def color_alternating(self, colors=['red', 'blue']):
        r'''Colors client items in alternating `colors`.

        Returns none.
        '''
        import abjad
        colors = abjad.CyclicTuple(colors)
        if isinstance(self._client, abjad.Component):
            target = [self._client]
        elif isinstance(self._client, abjad.selectiontools.LogicalTie):
            target = [self._client]
        elif isinstance(self._client, abjad.Selection):
            target = self._client
        else:
            raise TypeError(self._client)
        for i, item in enumerate(target):
            color = colors[i]
            abjad.label(item).color_leaves(color=color)

    def color_container(self, color='red'):
        r'''Colors contents of `container`.

        ..  container:: example

            Colors measure:

            ..  container:: example

                ::

                    >>> measure = Measure((2, 8), "c'8 d'8")
                    >>> label(measure).color_container('red')
                    >>> show(measure) # doctest: +SKIP

                ..  doctest::

                    >>> f(measure)
                    {
                        \override Accidental.color = #red
                        \override Beam.color = #red
                        \override Dots.color = #red
                        \override NoteHead.color = #red
                        \override Rest.color = #red
                        \override Stem.color = #red
                        \override TupletBracket.color = #red
                        \override TupletNumber.color = #red
                        \time 2/8
                        c'8
                        d'8
                        \revert Accidental.color
                        \revert Beam.color
                        \revert Dots.color
                        \revert NoteHead.color
                        \revert Rest.color
                        \revert Stem.color
                        \revert TupletBracket.color
                        \revert TupletNumber.color
                    }

            ..  container:: example expression

                ::

                    >>> measure = Measure((2, 8), "c'8 d'8")
                    >>> expression = label().color_container('red')
                    >>> expression(measure)
                    >>> show(measure) # doctest: +SKIP

                ..  doctest::

                    >>> f(measure)
                    {
                        \override Accidental.color = #red
                        \override Beam.color = #red
                        \override Dots.color = #red
                        \override NoteHead.color = #red
                        \override Rest.color = #red
                        \override Stem.color = #red
                        \override TupletBracket.color = #red
                        \override TupletNumber.color = #red
                        \time 2/8
                        c'8
                        d'8
                        \revert Accidental.color
                        \revert Beam.color
                        \revert Dots.color
                        \revert NoteHead.color
                        \revert Rest.color
                        \revert Stem.color
                        \revert TupletBracket.color
                        \revert TupletNumber.color
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        override(self.client).accidental.color = color
        override(self.client).beam.color = color
        override(self.client).dots.color = color
        override(self.client).note_head.color = color
        override(self.client).rest.color = color
        override(self.client).stem.color = color
        override(self.client).tuplet_bracket.color = color
        override(self.client).tuplet_number.color = color

    def color_leaves(self, color='red'):
        r"""Colors leaves.

        ..  container:: example

            Colors leaves red:

            ..  container:: example

                ::

                    >>> staff = Staff("cs'8. [ r8. s8. <c' cs' a'>8. ]")
                    >>> label(staff).color_leaves('red')
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        \once \override Accidental.color = #red
                        \once \override Beam.color = #red
                        \once \override Dots.color = #red
                        \once \override NoteHead.color = #red
                        \once \override Stem.color = #red
                        cs'8. [
                        \once \override Dots.color = #red
                        \once \override Rest.color = #red
                        r8.
                        s8.
                        \once \override Accidental.color = #red
                        \once \override Beam.color = #red
                        \once \override Dots.color = #red
                        \once \override NoteHead.color = #red
                        \once \override Stem.color = #red
                        <c' cs' a'>8. ]
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("cs'8. [ r8. s8. <c' cs' a'>8. ]")
                    >>> expression = label().color_leaves('red')
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        \once \override Accidental.color = #red
                        \once \override Beam.color = #red
                        \once \override Dots.color = #red
                        \once \override NoteHead.color = #red
                        \once \override Stem.color = #red
                        cs'8. [
                        \once \override Dots.color = #red
                        \once \override Rest.color = #red
                        r8.
                        s8.
                        \once \override Accidental.color = #red
                        \once \override Beam.color = #red
                        \once \override Dots.color = #red
                        \once \override NoteHead.color = #red
                        \once \override Stem.color = #red
                        <c' cs' a'>8. ]
                    }

        Returns none.
        """
        if self._expression:
            return self._update_expression(inspect.currentframe())
        for leaf in iterate(self.client).by_leaf():
            self._color_leaf(leaf, color)

    def color_note_heads(self, color_map=None):
        r'''Colors note note-heads.

        ..  container:: example

            Colors chord note-heads:

            ..  container:: example

                ::

                    >>> chord = Chord([12, 14, 18, 21, 23], (1, 4))
                    >>> pitches = [[-12, -10, 4], [-2, 8, 11, 17], [19, 27, 30, 33, 37]]
                    >>> colors = ['red', 'blue', 'green']
                    >>> color_map = pitchtools.NumberedPitchClassColorMap(pitches, colors)
                    >>> label(chord).color_note_heads(color_map)
                    >>> show(chord) # doctest: +SKIP

                ..  doctest::

                    >>> f(chord)
                    <
                        \tweak color #red
                        c''
                        \tweak color #red
                        d''
                        \tweak color #green
                        fs''
                        \tweak color #green
                        a''
                        \tweak color #blue
                        b''
                    >4

            ..  container:: example expression

                ::

                    >>> chord = Chord([12, 14, 18, 21, 23], (1, 4))
                    >>> pitches = [[-12, -10, 4], [-2, 8, 11, 17], [19, 27, 30, 33, 37]]
                    >>> colors = ['red', 'blue', 'green']
                    >>> color_map = pitchtools.NumberedPitchClassColorMap(pitches, colors)
                    >>> expression = label().color_note_heads(color_map)
                    >>> expression(chord)
                    >>> show(chord) # doctest: +SKIP

                ..  doctest::

                    >>> f(chord)
                    <
                        \tweak color #red
                        c''
                        \tweak color #red
                        d''
                        \tweak color #green
                        fs''
                        \tweak color #green
                        a''
                        \tweak color #blue
                        b''
                    >4

        ..  container:: example

            Colors note note-head:

            ..  container:: example

                ::

                    >>> note = Note("c'4")
                    >>> label(note).color_note_heads(color_map)
                    >>> show(note) # doctest: +SKIP

                ..  doctest::

                    >>> f(note)
                    \once \override NoteHead.color = #red
                    c'4

            ..  container:: example expression

                ::

                    >>> note = Note("c'4")
                    >>> expression = label().color_note_heads(color_map)
                    >>> expression(note)
                    >>> show(note) # doctest: +SKIP

                ..  doctest::

                    >>> f(note)
                    \once \override NoteHead.color = #red
                    c'4

        ..  container:: example

            Colors nothing:

            ..  container:: example

                ::

                    >>> staff = Staff()
                    >>> label(staff).color_note_heads(color_map)

            ..  container:: example expression

                ::

                    >>> staff = Staff()
                    >>> expression = label().color_note_heads(color_map)
                    >>> expression(staff)

        ..  container:: example

            Colors note-heads:

            ..  container:: example

                ::

                    >>> staff = Staff("c'8 cs'8 d'8 ds'8 e'8 f'8 fs'8 g'8 gs'8 a'8 as'8 b'8 c''8")
                    >>> label(staff).color_note_heads()
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        \once \override NoteHead.color = #(x11-color 'red)
                        c'8
                        \once \override NoteHead.color = #(x11-color 'MediumBlue)
                        cs'8
                        \once \override NoteHead.color = #(x11-color 'orange)
                        d'8
                        \once \override NoteHead.color = #(x11-color 'LightSlateBlue)
                        ds'8
                        \once \override NoteHead.color = #(x11-color 'ForestGreen)
                        e'8
                        \once \override NoteHead.color = #(x11-color 'MediumOrchid)
                        f'8
                        \once \override NoteHead.color = #(x11-color 'firebrick)
                        fs'8
                        \once \override NoteHead.color = #(x11-color 'DeepPink)
                        g'8
                        \once \override NoteHead.color = #(x11-color 'DarkOrange)
                        gs'8
                        \once \override NoteHead.color = #(x11-color 'IndianRed)
                        a'8
                        \once \override NoteHead.color = #(x11-color 'CadetBlue)
                        as'8
                        \once \override NoteHead.color = #(x11-color 'SeaGreen)
                        b'8
                        \once \override NoteHead.color = #(x11-color 'red)
                        c''8
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("c'8 cs'8 d'8 ds'8 e'8 f'8 fs'8 g'8 gs'8 a'8 as'8 b'8 c''8")
                    >>> expression = label().color_note_heads()
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        \once \override NoteHead.color = #(x11-color 'red)
                        c'8
                        \once \override NoteHead.color = #(x11-color 'MediumBlue)
                        cs'8
                        \once \override NoteHead.color = #(x11-color 'orange)
                        d'8
                        \once \override NoteHead.color = #(x11-color 'LightSlateBlue)
                        ds'8
                        \once \override NoteHead.color = #(x11-color 'ForestGreen)
                        e'8
                        \once \override NoteHead.color = #(x11-color 'MediumOrchid)
                        f'8
                        \once \override NoteHead.color = #(x11-color 'firebrick)
                        fs'8
                        \once \override NoteHead.color = #(x11-color 'DeepPink)
                        g'8
                        \once \override NoteHead.color = #(x11-color 'DarkOrange)
                        gs'8
                        \once \override NoteHead.color = #(x11-color 'IndianRed)
                        a'8
                        \once \override NoteHead.color = #(x11-color 'CadetBlue)
                        as'8
                        \once \override NoteHead.color = #(x11-color 'SeaGreen)
                        b'8
                        \once \override NoteHead.color = #(x11-color 'red)
                        c''8
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        color_map = color_map or self._pc_number_to_color
        for leaf in iterate(self.client).by_leaf():
            if isinstance(leaf, scoretools.Chord):
                for note_head in leaf.note_heads:
                    pc = note_head.written_pitch.numbered_pitch_class
                    color = color_map.get(pc, None)
                    if color is not None:
                        note_head.tweak.color = color
            elif isinstance(leaf, scoretools.Note):
                note_head = leaf.note_head
                pc = note_head.written_pitch.numbered_pitch_class
                color = color_map[pc.pitch_class_number]
                if color is not None:
                    override(leaf).note_head.color = color

    def remove_markup(self):
        r'''Removes markup from leaves.

        ..  container:: example

            Removes markup:

            ..  container:: example

                ::

                    >>> staff = Staff("c'8 d'8 e'8 f'8")
                    >>> label(staff).with_pitches()
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'8
                            ^ \markup {
                                \small
                                    c'
                                }
                        d'8
                            ^ \markup {
                                \small
                                    d'
                                }
                        e'8
                            ^ \markup {
                                \small
                                    e'
                                }
                        f'8
                            ^ \markup {
                                \small
                                    f'
                                }
                    }

                ::

                    >>> label(staff).remove_markup()
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'8
                        d'8
                        e'8
                        f'8
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("c'8 d'8 e'8 f'8")
                    >>> expression = label().with_pitches()
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'8
                            ^ \markup {
                                \small
                                    c'
                                }
                        d'8
                            ^ \markup {
                                \small
                                    d'
                                }
                        e'8
                            ^ \markup {
                                \small
                                    e'
                                }
                        f'8
                            ^ \markup {
                                \small
                                    f'
                                }
                    }

                ::

                    >>> expression = label().remove_markup()
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'8
                        d'8
                        e'8
                        f'8
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        leaves = iterate(self.client).by_leaf()
        for leaf in leaves:
            detach(markuptools.Markup, leaf)

    def vertical_moments(self, direction=Up, prototype=None):
        r'''Labels vertical moments.

        ..  container:: example

            Labels indices:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> label(staff_group).vertical_moments()
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        0
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        1
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        3
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        4
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        2
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> expression = label().vertical_moments()
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        0
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        1
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        3
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        4
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        2
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels pitch numbers:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedPitch
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                0
                                                -5
                                                -24
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                -5
                                                -24
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                4
                                                -7
                                                -24
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                -7
                                                -24
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                -7
                                                -24
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedPitch
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                0
                                                -5
                                                -24
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                -5
                                                -24
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                4
                                                -7
                                                -24
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                -7
                                                -24
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                -7
                                                -24
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels pitch-class numbers:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                7
                                                0
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                7
                                                2
                                                0
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                4
                                                0
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                0
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                2
                                                0
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                7
                                                0
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                7
                                                2
                                                0
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                4
                                                0
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                0
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                2
                                                0
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels interval numbers:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedInterval
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                15
                                                12
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                16
                                                12
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                17
                                                11
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                18
                                                11
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                16
                                                11
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedInterval
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                15
                                                12
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                16
                                                12
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                17
                                                11
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                18
                                                11
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                16
                                                11
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels interval-class numbers:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedIntervalClass
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                12
                                                7
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                7
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                4
                                                5
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                5
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                5
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.NumberedIntervalClass
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                12
                                                7
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                7
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                4
                                                5
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                5
                                                5
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \column
                                            {
                                                2
                                                5
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels interval-class vectors:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.IntervalClassVector
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            1000020
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0010020
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0100110
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            1000020
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0011010
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.IntervalClassVector
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            1000020
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0010020
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0100110
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            1000020
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \tiny
                                            0011010
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        ..  container:: example

            Labels set-classes:

            ..  container:: example

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.SetClass()
                    >>> label(staff_group).vertical_moments(prototype=prototype)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(2-5){0, 5}"
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-9){0, 2, 7}"
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-4){0, 1, 5}"
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(2-5){0, 5}"
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-7){0, 2, 5}"
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff_group = StaffGroup([])
                    >>> staff = Staff("c'8 d'4 e'16 f'16")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "alto" g4 f4""")
                    >>> staff_group.append(staff)
                    >>> staff = Staff(r"""\clef "bass" c,2""")
                    >>> staff_group.append(staff)
                    >>> prototype = pitchtools.SetClass()
                    >>> expression = label().vertical_moments(prototype=prototype)
                    >>> expression(staff_group)
                    >>> show(staff_group) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff_group)
                    \new StaffGroup <<
                        \new Staff {
                            c'8
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(2-5){0, 5}"
                                            }
                                    }
                            d'4
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-9){0, 2, 7}"
                                            }
                                    }
                            e'16
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-4){0, 1, 5}"
                                            }
                                    }
                            f'16
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(2-5){0, 5}"
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "alto"
                            g4
                            f4
                                ^ \markup {
                                    \tiny
                                        \line
                                            {
                                                "SC(3-7){0, 2, 5}"
                                            }
                                    }
                        }
                        \new Staff {
                            \clef "bass"
                            c,2
                        }
                    >>

        Set `prototype` to one of the classes shown above.

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        prototype = prototype or int
        vertical_moments = iterate(self.client).by_vertical_moment()
        for index, vertical_moment in enumerate(vertical_moments):
            label = None
            if prototype is int:
                label = markuptools.Markup(index, direction=direction)
            elif prototype is pitchtools.NumberedPitch:
                leaves = vertical_moment.leaves
                pitches = pitchtools.PitchSegment.from_selection(leaves)
                if not pitches:
                    continue
                pitch_numbers = [
                    pitch.numbered_pitch.pitch_number
                    for pitch in pitches
                    ]
                pitch_numbers = [markuptools.Markup(_) for _ in pitch_numbers]
                label = markuptools.Markup.column(pitch_numbers, direction)
            elif prototype is pitchtools.NumberedPitchClass:
                leaves = vertical_moment.leaves
                pitches = pitchtools.PitchSegment.from_selection(leaves)
                if not pitches:
                    continue
                pitch_classes = [
                    pitch.numbered_pitch_class.pitch_class_number
                    for pitch in pitches
                    ]
                pitch_classes = list(set(pitch_classes))
                pitch_classes.sort()
                pitch_classes.reverse()
                numbers = [str(_) for _ in pitch_classes]
                markup = [markuptools.Markup(_) for _ in numbers]
                label = markuptools.Markup.column(markup, direction)
            elif prototype is pitchtools.NumberedInterval:
                leaves = vertical_moment.leaves
                notes = [_ for _ in leaves if isinstance(_, scoretools.Note)]
                if not notes:
                    continue
                notes.sort(key=lambda x: x.written_pitch.numbered_pitch)
                notes.reverse()
                bass_note = notes[-1]
                upper_notes = notes[:-1]
                named_intervals = []
                for upper_note in upper_notes:
                    named_interval = pitchtools.NamedInterval.from_pitch_carriers(
                        bass_note.written_pitch, upper_note.written_pitch)
                    named_intervals.append(named_interval)
                numbers = [x.number for x in named_intervals]
                markup = [markuptools.Markup(_) for _ in numbers]
                label = markuptools.Markup.column(markup, direction)
            elif prototype is pitchtools.NumberedIntervalClass:
                leaves = vertical_moment.leaves
                notes = [_ for _ in leaves if isinstance(_, scoretools.Note)]
                if not notes:
                    continue
                notes.sort(key=lambda x: x.written_pitch.numbered_pitch)
                notes.reverse()
                bass_note = notes[-1]
                upper_notes = notes[:-1]
                numbers = []
                for upper_note in upper_notes:
                    interval = pitchtools.NamedInterval.from_pitch_carriers(
                        bass_note.written_pitch, upper_note.written_pitch)
                    interval_class = pitchtools.NumberedIntervalClass(interval)
                    number = interval_class.number
                    numbers.append(number)
                markup = [markuptools.Markup(_) for _ in numbers]
                label = markuptools.Markup.column(markup, direction)
            elif prototype is pitchtools.IntervalClassVector:
                leaves = vertical_moment.leaves
                pitches = pitchtools.PitchSegment.from_selection(leaves)
                if not pitches:
                    continue
                interval_class_vector = pitchtools.IntervalClassVector(
                    pitches,
                    item_class=pitchtools
                        .NumberedInversionEquivalentIntervalClass,
                    )
                markup = interval_class_vector._label
                label = markuptools.Markup(markup, direction=direction)
            elif (prototype is pitchtools.SetClass or
                isinstance(prototype, pitchtools.SetClass)):
                if prototype is pitchtools.SetClass:
                    prototype = prototype()
                assert isinstance(prototype, pitchtools.SetClass)
                leaves = vertical_moment.leaves
                pitch_class_set = pitchtools.PitchClassSet.from_selection(
                    leaves)
                if not pitch_class_set:
                    continue
                set_class = pitchtools.SetClass.from_pitch_class_set(
                    pitch_class_set,
                    lex_rank=prototype.lex_rank,
                    transposition_only=prototype.transposition_only,
                    )
                string = str(set_class)
                command = markuptools.MarkupCommand('line', [string])
                label = markuptools.Markup(command, direction=direction)
            else:
                message = 'unknown prototype: {!r}.'
                message = message.format(prototype)
                raise TypeError(message)
            if label is not None:
                label = label.tiny()
                if direction is Up:
                    leaf = vertical_moment.start_leaves[0]
                else:
                    leaf = vertical_moment.start_leaves[-1]
                attach(label, leaf)

    def with_durations(self, direction=Up, preferred_denominator=None):
        r'''Labels logical ties with durations.

        ..  container:: example

            Labels logical tie durations:

            ..  container:: example

                ::

                    >>> staff = Staff(r"c'4. d'8 ~ d'4. e'16 [ ef'16 ]")
                    >>> label(staff).with_durations()
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'4.
                            ^ \markup {
                                \small
                                    3/8
                                }
                        d'8 ~
                            ^ \markup {
                                \small
                                    1/2
                                }
                        d'4.
                        e'16 [
                            ^ \markup {
                                \small
                                    1/16
                                }
                        ef'16 ]
                            ^ \markup {
                                \small
                                    1/16
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff(r"c'4. d'8 ~ d'4. e'16 [ ef'16 ]")
                    >>> expression = label().with_durations()
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'4.
                            ^ \markup {
                                \small
                                    3/8
                                }
                        d'8 ~
                            ^ \markup {
                                \small
                                    1/2
                                }
                        d'4.
                        e'16 [
                            ^ \markup {
                                \small
                                    1/16
                                }
                        ef'16 ]
                            ^ \markup {
                                \small
                                    1/16
                                }
                    }

        ..  container:: example

            Labels logical ties with preferred denominator:

            ..  container:: example

                ::

                    >>> staff = Staff(r"c'4. d'8 ~ d'4. e'16 [ ef'16 ]")
                    >>> label(staff).with_durations(preferred_denominator=16)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'4.
                            ^ \markup {
                                \small
                                    6/16
                                }
                        d'8 ~
                            ^ \markup {
                                \small
                                    8/16
                                }
                        d'4.
                        e'16 [
                            ^ \markup {
                                \small
                                    1/16
                                }
                        ef'16 ]
                            ^ \markup {
                                \small
                                    1/16
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff(r"c'4. d'8 ~ d'4. e'16 [ ef'16 ]")
                    >>> expression = label().with_durations(preferred_denominator=16)
                    >>> expression(staff)
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff {
                        c'4.
                            ^ \markup {
                                \small
                                    6/16
                                }
                        d'8 ~
                            ^ \markup {
                                \small
                                    8/16
                                }
                        d'4.
                        e'16 [
                            ^ \markup {
                                \small
                                    1/16
                                }
                        ef'16 ]
                            ^ \markup {
                                \small
                                    1/16
                                }
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        logical_ties = iterate(self.client).by_logical_tie()
        for logical_tie in logical_ties:
            duration = logical_tie.get_duration()
            if preferred_denominator is not None:
                duration = mathtools.NonreducedFraction(duration)
                duration = duration.with_denominator(preferred_denominator)
            label = markuptools.Markup(str(duration), direction=direction)
            label = label.small()
            attach(label, logical_tie.head)

    def with_indices(self, direction=Up, prototype=None):
        r'''Labels logical ties with indices.

        ..  container:: example

            Labels logical tie indices:

            ..  container:: example

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> label(staff).with_indices()
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        af'8
                        gf'8 ~
                            ^ \markup {
                                \small
                                    3
                                }
                        gf'4
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> expression = label().with_indices()
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        af'8
                        gf'8 ~
                            ^ \markup {
                                \small
                                    3
                                }
                        gf'4
                    }

        ..  container:: example

            Labels note indices:

            ..  container:: example

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> label(staff).with_indices(prototype=Note)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                        <g' a'>4
                        af'8 ~
                            ^ \markup {
                                \small
                                    0
                                }
                        af'8
                            ^ \markup {
                                \small
                                    1
                                }
                        gf'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        gf'4
                            ^ \markup {
                                \small
                                    3
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> expression = label().with_indices(prototype=Note)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                        <g' a'>4
                        af'8 ~
                            ^ \markup {
                                \small
                                    0
                                }
                        af'8
                            ^ \markup {
                                \small
                                    1
                                }
                        gf'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        gf'4
                            ^ \markup {
                                \small
                                    3
                                }
                    }

        ..  container:: example

            Labels chord indices:

            ..  container:: example

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> label(staff).with_indices(prototype=Chord)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                        af'8
                        gf'8 ~
                        gf'4
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> expression = label().with_indices(prototype=Chord)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                        af'8
                        gf'8 ~
                        gf'4
                    }

        ..  container:: example

            Labels leaf indices:

            ..  container:: example

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> label(staff).with_indices(prototype=scoretools.Leaf)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        af'8
                            ^ \markup {
                                \small
                                    3
                                }
                        gf'8 ~
                            ^ \markup {
                                \small
                                    4
                                }
                        gf'4
                            ^ \markup {
                                \small
                                    5
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<c' bf'>8 <g' a'>4 af'8 ~ af'8 gf'8 ~ gf'4")
                    >>> expression = label().with_indices(prototype=scoretools.Leaf)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        <c' bf'>8
                            ^ \markup {
                                \small
                                    0
                                }
                        <g' a'>4
                            ^ \markup {
                                \small
                                    1
                                }
                        af'8 ~
                            ^ \markup {
                                \small
                                    2
                                }
                        af'8
                            ^ \markup {
                                \small
                                    3
                                }
                        gf'8 ~
                            ^ \markup {
                                \small
                                    4
                                }
                        gf'4
                            ^ \markup {
                                \small
                                    5
                                }
                    }

        ..  container:: example

            Labels tuplet indices:

            ..  container:: example

                ::

                    >>> staff = Staff(4 * Tuplet((2, 3), "c'8 [ d'8 e'8 ]"))
                    >>> label(staff).with_indices(prototype=Tuplet)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        0
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        1
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        2
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        3
                                    }
                            d'8
                            e'8
                        }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff(4 * Tuplet((2, 3), "c'8 [ d'8 e'8 ]"))
                    >>> expression = label().with_indices(prototype=Tuplet)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 2
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #2
                    } {
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        0
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        1
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        2
                                    }
                            d'8
                            e'8
                        }
                        \times 2/3 {
                            c'8
                                ^ \markup {
                                    \small
                                        3
                                    }
                            d'8
                            e'8
                        }
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        if prototype is None:
            items = iterate(self.client).by_logical_tie()
        else:
            items = iterate(self.client).by_class(prototype=prototype)
        items = list(items)
        for index, item in enumerate(items):
            string = str(index)
            label = markuptools.Markup(string, direction=direction)
            label = label.small()
            leaves = iterate(item).by_leaf()
            first_leaf = list(leaves)[0]
            attach(label, first_leaf)

    def with_intervals(self, direction=Up, prototype=None):
        r"""Labels consecutive notes with intervals.

        ..  container:: example

            Labels consecutive notes with interval names:

            ..  container:: example

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> label(staff).with_intervals(prototype=None)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +aug15 }
                        cs'''4 ^ \markup { -M9 }
                        b'4 ^ \markup { -aug9 }
                        af4 ^ \markup { -m7 }
                        bf,4 ^ \markup { +aug1 }
                        b,4 ^ \markup { +m14 }
                        a'4 ^ \markup { +m2 }
                        bf'4
                    }

            ..  container:: example expression

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> expression = label().with_intervals(prototype=None)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +aug15 }
                        cs'''4 ^ \markup { -M9 }
                        b'4 ^ \markup { -aug9 }
                        af4 ^ \markup { -m7 }
                        bf,4 ^ \markup { +aug1 }
                        b,4 ^ \markup { +m14 }
                        a'4 ^ \markup { +m2 }
                        bf'4
                    }

        ..  container:: example

            Labels consecutive notes with interval-class names:

            ..  container:: example

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NamedIntervalClass
                    >>> label(staff).with_intervals(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +aug8 }
                        cs'''4 ^ \markup { -M2 }
                        b'4 ^ \markup { -aug2 }
                        af4 ^ \markup { -m7 }
                        bf,4 ^ \markup { aug1 }
                        b,4 ^ \markup { +m7 }
                        a'4 ^ \markup { +m2 }
                        bf'4
                    }

            ..  container:: example expression

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NamedIntervalClass
                    >>> expression = label().with_intervals(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +aug8 }
                        cs'''4 ^ \markup { -M2 }
                        b'4 ^ \markup { -aug2 }
                        af4 ^ \markup { -m7 }
                        bf,4 ^ \markup { aug1 }
                        b,4 ^ \markup { +m7 }
                        a'4 ^ \markup { +m2 }
                        bf'4
                    }

        ..  container:: example

            Labels consecutive notes with interval numbers:

            ..  container:: example

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedInterval
                    >>> label(staff).with_intervals(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +25 }
                        cs'''4 ^ \markup { -14 }
                        b'4 ^ \markup { -15 }
                        af4 ^ \markup { -10 }
                        bf,4 ^ \markup { +1 }
                        b,4 ^ \markup { +22 }
                        a'4 ^ \markup { +1 }
                        bf'4
                    }

            ..  container:: example expression

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedInterval
                    >>> expression = label().with_intervals(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +25 }
                        cs'''4 ^ \markup { -14 }
                        b'4 ^ \markup { -15 }
                        af4 ^ \markup { -10 }
                        bf,4 ^ \markup { +1 }
                        b,4 ^ \markup { +22 }
                        a'4 ^ \markup { +1 }
                        bf'4
                    }


        ..  container:: example

            Labels consecutive notes with interval-class numbers:

            ..  container:: example

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedIntervalClass
                    >>> label(staff).with_intervals(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +1 }
                        cs'''4 ^ \markup { -2 }
                        b'4 ^ \markup { -3 }
                        af4 ^ \markup { -10 }
                        bf,4 ^ \markup { +1 }
                        b,4 ^ \markup { +10 }
                        a'4 ^ \markup { +1 }
                        bf'4
                    }

            ..  container:: example expression

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedIntervalClass
                    >>> expression = label().with_intervals(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { +1 }
                        cs'''4 ^ \markup { -2 }
                        b'4 ^ \markup { -3 }
                        af4 ^ \markup { -10 }
                        bf,4 ^ \markup { +1 }
                        b,4 ^ \markup { +10 }
                        a'4 ^ \markup { +1 }
                        bf'4
                    }

        ..  container:: example

            Labels consecutive notes with inversion-equivalent interval-class
            numbers:

            ..  container:: example

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedInversionEquivalentIntervalClass
                    >>> label(staff).with_intervals(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { 1 }
                        cs'''4 ^ \markup { 2 }
                        b'4 ^ \markup { 3 }
                        af4 ^ \markup { 2 }
                        bf,4 ^ \markup { 1 }
                        b,4 ^ \markup { 2 }
                        a'4 ^ \markup { 1 }
                        bf'4
                    }

            ..  container:: example expression

                ::

                    >>> pitch_numbers = [0, 25, 11, -4, -14, -13, 9, 10]
                    >>> notes = scoretools.make_notes(pitch_numbers, [(1, 4)])
                    >>> staff = Staff(notes)
                    >>> prototype = pitchtools.NumberedInversionEquivalentIntervalClass
                    >>> expression = label().with_intervals(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        c'4 ^ \markup { 1 }
                        cs'''4 ^ \markup { 2 }
                        b'4 ^ \markup { 3 }
                        af4 ^ \markup { 2 }
                        bf,4 ^ \markup { 1 }
                        b,4 ^ \markup { 2 }
                        a'4 ^ \markup { 1 }
                        bf'4
                    }

        Returns none.
        """
        if self._expression:
            return self._update_expression(inspect.currentframe())
        prototype = prototype or pitchtools.NamedInterval
        notes = iterate(self.client).by_class(scoretools.Note)
        for note in notes:
            label = None
            next_leaf = inspect_(note).get_leaf(1)
            if isinstance(next_leaf, scoretools.Note):
                interval = pitchtools.NamedInterval.from_pitch_carriers(
                    note,
                    next_leaf,
                    )
                if prototype is pitchtools.NamedInterval:
                    label = markuptools.Markup(interval, direction)
                elif prototype is pitchtools.NamedIntervalClass:
                    label = pitchtools.NamedIntervalClass(interval)
                    label = markuptools.Markup(label, direction)
                elif prototype is pitchtools.NumberedInterval:
                    label = pitchtools.NumberedInterval(interval)
                    label = markuptools.Markup(label, direction)
                elif prototype is pitchtools.NumberedIntervalClass:
                    label = pitchtools.NumberedIntervalClass(interval)
                    label = markuptools.Markup(label, direction)
                elif prototype is pitchtools.NumberedInversionEquivalentIntervalClass:
                    label = pitchtools.NumberedInversionEquivalentIntervalClass(interval)
                    label = markuptools.Markup(label, direction)
                if label is not None:
                    attach(label, note)

    def with_pitches(self, direction=Up, prototype=None):
        r'''Labels logical ties with pitches.

        ..  container:: example

            Labels logical ties with pitch names:

            ..  container:: example

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> label(staff).with_pitches(prototype=None)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            fs'
                                            d'
                                            a
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    g'
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    fs''
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> expression = label().with_pitches(prototype=None)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            fs'
                                            d'
                                            a
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    g'
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    fs''
                                }
                    }

        ..  container:: example

            Labels logical ties with pitch numbers:

            ..  container:: example

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> prototype = pitchtools.NumberedPitch
                    >>> label(staff).with_pitches(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            6
                                            2
                                            -3
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    7
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    18
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> prototype = pitchtools.NumberedPitch
                    >>> expression = label().with_pitches(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            6
                                            2
                                            -3
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    7
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    18
                                }
                    }

        ..  container:: example

            Labels logical ties with pitch-class numbers:

            ..  container:: example

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> label(staff).with_pitches(prototype=prototype)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            6
                                            2
                                            9
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    7
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    6
                                }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff("<a d' fs'>4 g'4 ~ g'8 r8 fs''4")
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> expression = label().with_pitches(prototype=prototype)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                    } {
                        <a d' fs'>4
                            ^ \markup {
                                \small
                                    \column
                                        {
                                            6
                                            2
                                            9
                                        }
                                }
                        g'4 ~
                            ^ \markup {
                                \small
                                    7
                                }
                        g'8
                        r8
                        fs''4
                            ^ \markup {
                                \small
                                    6
                                }
                    }

        ..  container:: example

            Labels logical ties with pitch names (filtered by selection):

            ..  container:: example

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> label(selections).with_pitches()
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    df''
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    c''
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    d''
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    ds''
                                }
                    }

            ..  container:: example expression

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> expression = label().with_pitches()
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    df''
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    c''
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    d''
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    ds''
                                }
                    }

        ..  container:: example

            Labels logical ties with pitch numbers (filtered by selection):

            ..  container:: example

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.NumberedPitch
                    >>> label(selections).with_pitches(prototype=prototype)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    13
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    12
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    14
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    15
                                }
                    }

            ..  container:: example expression

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.NumberedPitch
                    >>> expression = label().with_pitches(prototype=prototype)
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    13
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    12
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    14
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    15
                                }
                    }

        ..  container:: example

            Labels logical ties with pitch-class numbers (filtered by
            selection):

            ..  container:: example

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> label(selections).with_pitches(prototype=prototype)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    1
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    0
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    2
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    3
                                }
                    }

            ..  container:: example expression

                ::

                    >>> voice = Voice("df''4 c''4 f'4 fs'4 d''4 ds''4")
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:2], voice[-2:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.NumberedPitchClass
                    >>> expression = label().with_pitches(prototype=prototype)
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''4 \startGroup
                            ^ \markup {
                                \small
                                    1
                                }
                        c''4 \stopGroup
                            ^ \markup {
                                \small
                                    0
                                }
                        f'4
                        fs'4
                        d''4 \startGroup
                            ^ \markup {
                                \small
                                    2
                                }
                        ds''4 \stopGroup
                            ^ \markup {
                                \small
                                    3
                                }
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        prototype = prototype or pitchtools.NamedPitch
        logical_ties = iterate(self.client).by_logical_tie()
        for logical_tie in logical_ties:
            leaf = logical_tie.head
            label = None
            if prototype is pitchtools.NamedPitch:
                if isinstance(leaf, scoretools.Note):
                    pitch = leaf.written_pitch
                    label = str(pitch)
                    label = markuptools.Markup(label, direction=direction)
                    label = label.small()
                elif isinstance(leaf, scoretools.Chord):
                    pitches = leaf.written_pitches
                    pitches = reversed(pitches)
                    pitches = [markuptools.Markup(_) for _ in pitches]
                    label = markuptools.Markup.column(pitches)
                    label = label.small()
            elif prototype is pitchtools.NumberedPitch:
                if isinstance(leaf, scoretools.Note):
                    pitch = leaf.written_pitch.pitch_number
                    label = str(pitch)
                    label = markuptools.Markup(label, direction=direction)
                    label = label.small()
                elif isinstance(leaf, scoretools.Chord):
                    pitches = leaf.written_pitches
                    pitches = reversed(pitches)
                    pitches = [_.pitch_number for _ in pitches]
                    pitches = [markuptools.Markup(_) for _ in pitches]
                    label = markuptools.Markup.column(pitches)
                    label = label.small()
            elif prototype is pitchtools.NumberedPitchClass:
                if isinstance(leaf, scoretools.Note):
                    pitch = leaf.written_pitch.pitch_class_number
                    label = str(pitch)
                    label = markuptools.Markup(label, direction=direction)
                    label = label.small()
                elif isinstance(leaf, scoretools.Chord):
                    pitches = leaf.written_pitches
                    pitches = reversed(pitches)
                    pitches = [_.pitch_class_number for _ in pitches]
                    pitches = [markuptools.Markup(_) for _ in pitches]
                    label = markuptools.Markup.column(pitches)
                    label = label.small()
            if label is not None:
                label = new(label, direction=direction)
                attach(label, leaf)

    def with_set_classes(self, direction=Up, prototype=None):
        r'''Labels items in client with set-classes.

        ..  container:: example

            Labels selections with Forte-ranked transposition-inversion
            set-classes:

            ..  container:: example

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> label(selections).with_set_classes()
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-3){0, 1, 3, 4}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-20){0, 1, 5, 8}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

            ..  container:: example expression

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> expression = label().with_set_classes()
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-3){0, 1, 3, 4}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-20){0, 1, 5, 8}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

        ..  container:: example

            Labels selections with lex-ranked transposition-inversion
            set-classes:

            ..  container:: example

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.SetClass(lex_rank=True)
                    >>> label(selections).with_set_classes(prototype=prototype)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-6){0, 1, 3, 4}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-16){0, 1, 5, 8}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

            ..  container:: example expression

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.SetClass(lex_rank=True)
                    >>> expression = label().with_set_classes(prototype=prototype)
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-6){0, 1, 3, 4}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-16){0, 1, 5, 8}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

        ..  container:: example

            Labels selections with transposition-only set-classes:

            ..  container:: example

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.SetClass(transposition_only=True)
                    >>> label(selections).with_set_classes(prototype=prototype)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-3){0, 1, 2, 5}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-20){0, 2, 3, 6}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

            ..  container:: example expression

                ::

                    >>> string = "df''8 c''8 bf'8 a'8 f'4. fs'8 g'8 b'8 d''2."
                    >>> voice = Voice(string)
                    >>> voice.consists_commands.append('Horizontal_bracket_engraver')
                    >>> selections = [voice[:4], voice[-4:]]
                    >>> for selection in selections:
                    ...     spanner = spannertools.HorizontalBracketSpanner()
                    ...     attach(spanner, selection)
                    ...
                    >>> prototype = pitchtools.SetClass(transposition_only=True)
                    >>> expression = label().with_set_classes(prototype=prototype)
                    >>> expression(selections)
                    >>> override(voice).horizontal_bracket.staff_padding = 3
                    >>> override(voice).text_script.staff_padding = 2
                    >>> show(voice) # doctest: +SKIP

                ..  doctest::

                    >>> f(voice)
                    \new Voice \with {
                        \consists Horizontal_bracket_engraver
                        \override HorizontalBracket.staff-padding = #3
                        \override TextScript.staff-padding = #2
                    } {
                        df''8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-3){0, 1, 2, 5}"
                                        }
                                }
                        c''8
                        bf'8
                        a'8 \stopGroup
                        f'4.
                        fs'8 \startGroup
                            ^ \markup {
                                \tiny
                                    \line
                                        {
                                            "SC(4-20){0, 2, 3, 6}"
                                        }
                                }
                        g'8
                        b'8
                        d''2. \stopGroup
                    }

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        prototype = prototype or pitchtools.SetClass()
        if prototype is pitchtools.SetClass:
            prototype = prototype()
        assert isinstance(prototype, pitchtools.SetClass), repr(prototype)
        for selection in self._client:
            pitch_class_set = pitchtools.PitchClassSet.from_selection(
                selection)
            if not pitch_class_set:
                continue
            set_class = pitchtools.SetClass.from_pitch_class_set(
                pitch_class_set,
                lex_rank=prototype.lex_rank,
                transposition_only=prototype.transposition_only,
                )
            string = str(set_class)
            command = markuptools.MarkupCommand('line', [string])
            label = markuptools.Markup(command, direction=direction)
            if label is not None:
                label = label.tiny()
                leaf = selection[0]
                attach(label, leaf)

    def with_start_offsets(
        self,
        clock_time=False,
        direction=Up,
        font_size=None,
        ):
        r'''Labels logical ties with start offsets.

        ..  container:: example

            Labels logical tie start offsets:

            ..  container:: example

                ::

                    >>> staff = Staff(r"\times 2/3 { c'4 d'4 e'4 ~ } e'4 ef'4")
                    >>> label(staff).with_start_offsets(direction=Up)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                        \override TupletBracket.staff-padding = #0
                    } {
                        \times 2/3 {
                            c'4 ^ \markup { 0 }
                            d'4 ^ \markup { 1/6 }
                            e'4 ~ ^ \markup { 1/3 }
                        }
                        e'4
                        ef'4 ^ \markup { 3/4 }
                    }

            ..  container:: example expression

                ::

                    >>> staff = Staff(r"\times 2/3 { c'4 d'4 e'4 ~ } e'4 ef'4")
                    >>> expression = label().with_start_offsets(direction=Up)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(staff) # doctest: +SKIP

                ..  doctest::

                    >>> f(staff)
                    \new Staff \with {
                        \override TextScript.staff-padding = #4
                        \override TupletBracket.staff-padding = #0
                    } {
                        \times 2/3 {
                            c'4 ^ \markup { 0 }
                            d'4 ^ \markup { 1/6 }
                            e'4 ~ ^ \markup { 1/3 }
                        }
                        e'4
                        ef'4 ^ \markup { 3/4 }
                    }

        ..  container:: example

            Labels logical tie start offsets with clock time:

            ..  container:: example

                ::

                    >>> staff = Staff(r"c'2 d' e' f'")
                    >>> score = Score([staff])
                    >>> attach(Tempo(Duration(1, 4), 60), staff[0])
                    >>> label(staff).with_start_offsets(clock_time=True)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(score) # doctest: +SKIP

                ..  doctest::

                    >>> f(score)
                    \new Score <<
                        \new Staff \with {
                            \override TextScript.staff-padding = #4
                            \override TupletBracket.staff-padding = #0
                        } {
                            \tempo 4=60
                            c'2 ^ \markup { 0'00'' }
                            d'2 ^ \markup { 0'02'' }
                            e'2 ^ \markup { 0'04'' }
                            f'2 ^ \markup { 0'06'' }
                        }
                    >>

            ..  container:: example

                ::

                    >>> staff = Staff(r"c'2 d' e' f'")
                    >>> score = Score([staff])
                    >>> attach(Tempo(Duration(1, 4), 60), staff[0])
                    >>> expression = label().with_start_offsets(clock_time=True)
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(score) # doctest: +SKIP

                ..  doctest::

                    >>> f(score)
                    \new Score <<
                        \new Staff \with {
                            \override TextScript.staff-padding = #4
                            \override TupletBracket.staff-padding = #0
                        } {
                            \tempo 4=60
                            c'2 ^ \markup { 0'00'' }
                            d'2 ^ \markup { 0'02'' }
                            e'2 ^ \markup { 0'04'' }
                            f'2 ^ \markup { 0'06'' }
                        }
                    >>

        ..  container:: example

            Labels logical tie start offsets with clock time and custom font
            size:

            ..  container:: example

                ::

                    >>> staff = Staff(r"c'2 d' e' f'")
                    >>> score = Score([staff])
                    >>> attach(Tempo(Duration(1, 4), 60), staff[0])
                    >>> label(staff).with_start_offsets(
                    ...     clock_time=True,
                    ...     font_size=-3,
                    ...     )
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(score) # doctest: +SKIP

                ..  doctest::

                    >>> f(score)
                    \new Score <<
                        \new Staff \with {
                            \override TextScript.staff-padding = #4
                            \override TupletBracket.staff-padding = #0
                        } {
                            \tempo 4=60
                            c'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'00''
                                    }
                            d'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'02''
                                    }
                            e'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'04''
                                    }
                            f'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'06''
                                    }
                        }
                    >>

            ..  container:: example expression

                ::

                    >>> staff = Staff(r"c'2 d' e' f'")
                    >>> score = Score([staff])
                    >>> attach(Tempo(Duration(1, 4), 60), staff[0])
                    >>> expression = label().with_start_offsets(
                    ...     clock_time=True,
                    ...     font_size=-3,
                    ...     )
                    >>> expression(staff)
                    >>> override(staff).text_script.staff_padding = 4
                    >>> override(staff).tuplet_bracket.staff_padding = 0
                    >>> show(score) # doctest: +SKIP

                ..  doctest::

                    >>> f(score)
                    \new Score <<
                        \new Staff \with {
                            \override TextScript.staff-padding = #4
                            \override TupletBracket.staff-padding = #0
                        } {
                            \tempo 4=60
                            c'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'00''
                                    }
                            d'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'02''
                                    }
                            e'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'04''
                                    }
                            f'2
                                ^ \markup {
                                    \fontsize
                                        #-3
                                        0'06''
                                    }
                        }
                    >>

        Returns none.
        '''
        if self._expression:
            return self._update_expression(inspect.currentframe())
        logical_ties = iterate(self.client).by_logical_tie()
        for logical_tie in logical_ties:
            if clock_time:
                inspector = inspect_(logical_tie.head)
                timespan = inspector.get_timespan(in_seconds=True)
                start_offset = timespan.start_offset
                string = start_offset.to_clock_string()
                string = '"{}"'.format(string)
            else:
                timespan = inspect_(logical_tie.head).get_timespan()
                start_offset = timespan.start_offset
                string = str(start_offset)
            label = markuptools.Markup(string, direction=direction)
            if font_size is not None:
                label = label.fontsize(font_size)
            attach(label, logical_tie.head)



# -*- coding: utf-8 -*-
from abjad import *


def test_timespantools_Timespan___and___01():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(-10, -5)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([])

def test_timespantools_Timespan___and___02():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(-10, 0)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([])

def test_timespantools_Timespan___and___03():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(-10, 5)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 5)
    ])

def test_timespantools_Timespan___and___04():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(-10, 15)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 15)
    ])

def test_timespantools_Timespan___and___05():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(-10, 25)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 15)
    ])

def test_timespantools_Timespan___and___06():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(0, 10)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 10)
    ])

def test_timespantools_Timespan___and___07():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(0, 15)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 15)
    ])

def test_timespantools_Timespan___and___08():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(5, 10)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(5, 10)
    ])

def test_timespantools_Timespan___and___09():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(5, 15)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(5, 15)
    ])

def test_timespantools_Timespan___and___10():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(0, 25)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(0, 15)
    ])

def test_timespantools_Timespan___and___11():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(5, 25)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([
        timespantools.Timespan(5, 15)
    ])

def test_timespantools_Timespan___and___12():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(15, 25)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([])

def test_timespantools_Timespan___and___13():
    timespan_1 = timespantools.Timespan(0, 15)
    timespan_2 = timespantools.Timespan(20, 25)
    result = timespan_1 & timespan_2
    assert result == timespantools.TimespanList([])



# -*- coding: utf-8 -*-
from abjad.tools.abctools import AbjadValueObject
from abjad.tools.topleveltools import select


class ContiguitySelectorCallback(AbjadValueObject):
    r'''Contiguity selector callback.
    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Callbacks'

    __slots__ = (
        )

    ### SPECIAL METHODS ###

    def __call__(self, argument, rotation=None):
        r'''Iterates tuple `argument`.
        '''
        result = []
        subresult = []
        subresult.extend(argument[:1])
        for subexpr in argument[1:]:
            try:
                that_timespan = subresult[-1]._get_timespan()
            except AttributeError:
                that_timespan = subresult[-1].get_timespan()
            try:
                this_timespan = subexpr._get_timespan()
            except AttributeError:
                this_timespan = subexpr.get_timespan()
            if that_timespan.stop_offset == this_timespan.start_offset:
                subresult.append(subexpr)
            else:
                result.append(select(subresult))
                subresult = [subexpr]
        if subresult:
            result.append(select(subresult))
        return tuple(result)



# -*- coding: utf-8 -*-
from abjad import *


def test_selectiontools_VerticalMoment___init___01():
    r'''Initializes vertical moment from empty input.
    '''

    vertical_moment = selectiontools.VerticalMoment()

    assert repr(vertical_moment) == 'VerticalMoment()'



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools.pitchtools import PitchClassSet


def test_pitchtools_PitchClassSet___repr___01():
    r'''Named pitch-class set repr is evaluable.
    '''

    named_pitch_classes = ['gs', 'a', 'as', 'c', 'cs']
    named_pitch_class_set_1 = pitchtools.PitchClassSet(named_pitch_classes)
    named_pitch_class_set_2 = eval(repr(named_pitch_class_set_1))

    "PitchClassSet(['a', 'as', 'c', 'cs', 'gs'])"

    assert isinstance(named_pitch_class_set_1, pitchtools.PitchClassSet)
    assert isinstance(named_pitch_class_set_2, pitchtools.PitchClassSet)


def test_pitchtools_PitchClassSet___repr___02():

    numbered_pitch_class_set_1 = pitchtools.PitchClassSet(
        [6, 7, 10, 10.5])
    numbered_pitch_class_set_2 = eval(repr(numbered_pitch_class_set_1))

    assert isinstance(numbered_pitch_class_set_1,
        pitchtools.PitchClassSet)
    assert isinstance(numbered_pitch_class_set_2,
        pitchtools.PitchClassSet)



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_make_percussion_note_01():
    r'''tied total_duration < max_note_duration.
    '''

    note = scoretools.make_percussion_note(1, (5, 64), (1, 1))

    assert len(note) == 2
    assert note[0].written_pitch.numbered_pitch == 1
    assert isinstance(note[0], Note)
    assert isinstance(note[1], Rest)
    assert note[0].written_duration == Duration(1, 16)
    assert note[1].written_duration == Duration(1, 64)
    assert all(len(inspect_(x).get_logical_tie()) == 1 for x in note)


def test_scoretools_make_percussion_note_02():
    r'''max_note_duration < tied total_duration.
    '''

    note = scoretools.make_percussion_note(1, (5, 64), (1, 64))

    assert len(note) == 2
    assert isinstance(note[0], Note)
    assert isinstance(note[1], Rest)
    assert note[0].written_duration == Duration(1, 64)
    assert note[1].written_duration == Duration(1, 16)
    assert all(len(inspect_(x).get_logical_tie()) == 1 for x in note)


def test_scoretools_make_percussion_note_03():
    r'''non-tied total_duration < max_note_duration.
    '''

    note = scoretools.make_percussion_note(1, (3, 64), (1, 1))

    assert len(note) == 1
    assert isinstance(note[0], Note)
    assert note[0].written_duration == Duration(3, 64)


def test_scoretools_make_percussion_note_04():
    r'''max_note_duration < non-tied total_duration.
    '''

    t = scoretools.make_percussion_note(1, (3, 64), (1, 64))

    assert len(t) == 2
    assert isinstance(t[0], Note)
    assert isinstance(t[1], Rest)
    assert t[0].written_duration == Duration(1, 64)
    assert t[1].written_duration == Duration(1, 32)
    assert all(len(inspect_(x).get_logical_tie()) == 1 for x in t)



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_Score_add_final_markup_01():

    staff = Staff("c'4 d'4 e'4 f'4")
    score = Score([staff])
    markup = r'\italic \right-column { "Bremen - Boston - Los Angeles." "Jul 2010 - May 2011." }'
    markup = markuptools.Markup(markup, Down)
    score.add_final_markup(markup, extra_offset=(4, -2))


    r'''
    \new Score <<
        \new Staff {
            c'4
            d'4
            e'4
            \once \override TextScript.extra-offset = #'(4 . -2)
            f'4
                _ \markup {
                    \italic
                        \right-column
                            {
                                "Bremen - Boston - Los Angeles."
                                "Jul 2010 - May 2011."
                            }
                    }
        }
    >>
    '''

    assert format(score) == stringtools.normalize(
        r'''
        \new Score <<
            \new Staff {
                c'4
                d'4
                e'4
                \once \override TextScript.extra-offset = #'(4 . -2)
                f'4
                    _ \markup {
                        \italic
                            \right-column
                                {
                                    "Bremen - Boston - Los Angeles."
                                    "Jul 2010 - May 2011."
                                }
                        }
            }
        >>
        '''
        )



# -*- coding: utf-8 -*-
import copy
from abjad import *


def test_pitchtools_NamedPitch___copy___01():

    pitch = NamedPitch(13)
    new = copy.copy(pitch)

    assert new is not pitch
    assert new.accidental is not pitch.accidental



# -*- coding: utf-8 -*-
import copy
from abjad import Beam
from abjad import Clef
from abjad import Crescendo
from abjad import Decrescendo
from abjad import Dynamic
from abjad import Markup
from abjad import Measure
from abjad import Score
from abjad import Slur
from abjad import Staff
from abjad import StaffGroup
from abjad import Tie
from abjad import Voice
from abjad import attach
from abjad import indicatortools
from abjad import override
from abjad import spannertools
from abjad.tools.topleveltools import select


def make_bartok_score():
    score = Score([])
    piano_staff = StaffGroup([], context_name='PianoStaff')
    upper_staff = Staff([])
    lower_staff = Staff([])
    piano_staff.append(upper_staff)
    piano_staff.append(lower_staff)
    score.append(piano_staff)
    upper_measures = []
    upper_measures.append(Measure((2, 4), []))
    upper_measures.append(Measure((3, 4), []))
    upper_measures.append(Measure((2, 4), []))
    upper_measures.append(Measure((2, 4), []))
    upper_measures.append(Measure((2, 4), []))
    lower_measures = copy.deepcopy(upper_measures)
    upper_staff.extend(upper_measures)
    lower_staff.extend(lower_measures)
    upper_measures[0].extend("a'8 g'8 f'8 e'8")
    upper_measures[1].extend("d'4 g'8 f'8 e'8 d'8")
    upper_measures[2].extend("c'8 d'16 e'16 f'8 e'8")
    upper_measures[3].append("d'2")
    upper_measures[4].append("d'2")
    lower_measures[0].extend("b4 d'8 c'8")
    lower_measures[1].extend("b8 a8 af4 c'8 bf8")
    lower_measures[2].extend("a8 g8 fs8 g16 a16")
    upper_voice = Voice("b2", name='upper voice')
    command = indicatortools.LilyPondCommand('voiceOne')
    attach(command, upper_voice)
    lower_voice = Voice("b4 a4", name='lower voice')
    command = indicatortools.LilyPondCommand('voiceTwo')
    attach(command, lower_voice)
    lower_measures[3].extend([upper_voice, lower_voice])
    lower_measures[3].is_simultaneous = True
    upper_voice = Voice("b2", name='upper voice')
    command = indicatortools.LilyPondCommand('voiceOne')
    attach(command, upper_voice)
    lower_voice = Voice("g2", name='lower voice')
    command = indicatortools.LilyPondCommand('voiceTwo')
    attach(command, lower_voice)
    lower_measures[4].extend([upper_voice, lower_voice])
    lower_measures[4].is_simultaneous = True
    clef = Clef('bass')
    attach(clef, lower_staff)
    dynamic = Dynamic('pp')
    attach(dynamic, upper_measures[0][0])
    dynamic = Dynamic('mp')
    attach(dynamic, upper_measures[1][1])
    dynamic = Dynamic('pp')
    attach(dynamic, lower_measures[0][1])
    dynamic = Dynamic('mp')
    attach(dynamic, lower_measures[1][3])
    score.add_final_bar_line()
    selector = select().by_leaf(flatten=True)
    upper_leaves = selector(upper_staff)
    lower_leaves = selector(lower_staff)
    beam = Beam()
    attach(beam, upper_leaves[:4])
    beam = Beam()
    attach(beam, lower_leaves[1:5])
    beam = Beam()
    attach(beam, lower_leaves[6:10])
    slur = Slur()
    attach(slur, upper_leaves[:5])
    slur = Slur()
    attach(slur, upper_leaves[5:])
    slur = Slur()
    attach(slur, lower_leaves[1:6])
    crescendo = Crescendo()
    attach(crescendo, upper_leaves[-7:-2])
    decrescendo = Decrescendo()
    attach(decrescendo, upper_leaves[-2:])
    markup = Markup('ritard.')
    text_spanner = spannertools.TextSpanner()
    override(text_spanner).text_spanner.bound_details__left__text = markup
    attach(text_spanner, upper_leaves[-7:])
    tie = Tie()
    attach(tie, upper_leaves[-2:])
    note_1 = lower_staff[-2]['upper voice'][0]
    note_2 = lower_staff[-1]['upper voice'][0]
    notes = [note_1, note_2]
    tie = Tie()
    attach(tie, notes)
    return score


# -*- coding: utf-8 -*-
from __future__ import print_function
import collections
import os
from abjad.tools import abctools
from abjad.tools import systemtools


class LaTeXDocumentHandler(abctools.AbjadObject):
    r"""A LaTeX document handler.

    ::

        >>> input_file_contents = '''Let's print something:
        ...
        ... \\begin{comment}
        ... <abjad>
        ... print("hello, world!")
        ... </abjad>
        ... \\end{comment}
        ...
        ... This is just a simple Python string:
        ...
        ... \\begin{comment}
        ... <abjad>
        ... just_a_string = \'\'\'
        ... show(Nothing!)
        ... \'\'\'
        ... </abjad>
        ... \\end{comment}
        ...
        ... And let's show some music too:
        ...
        ... \\begin{comment}
        ... <abjad>
        ... show(Note("c'4"))
        ... </abjad>
        ... \\end{comment}
        ...
        ... That's it!
        ... '''

    ::

        >>> from abjad.tools import abjadbooktools
        >>> document_handler = abjadbooktools.LaTeXDocumentHandler(
        ...     input_file_contents=input_file_contents,
        ...     input_file_path='test.tex.raw',
        ...     assets_directory='images',
        ...     )

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Document Handlers'

    __slots__ = (
        '_assets_directory',
        '_errored',
        '_input_file_contents',
        '_input_file_path',
        '_latex_root_directory',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        assets_directory=None,
        input_file_contents=None,
        input_file_path=None,
        latex_root_directory=None,
        ):
        self._assets_directory = assets_directory
        self._errored = False
        input_file_contents = input_file_contents or ()
        if isinstance(input_file_contents, str):
            input_file_contents = input_file_contents.rstrip().splitlines()
        self._input_file_contents = tuple(input_file_contents)
        self._input_file_path = input_file_path
        self._latex_root_directory = latex_root_directory

    ### SPECIAL METHODS ###

    def __call__(
        self,
        clean=None,
        configuration=None,
        output_file_path=None,
        return_source=None,
        skip_rendering=None,
        stylesheet=None,
        verbose=None,
        ):
        import abjad
        from abjad.tools import abjadbooktools
        configuration = configuration or {}
        if output_file_path:
            output_file_path = os.path.abspath(output_file_path)
            _, output_file_path = os.path.split(output_file_path)
            output_directory, _ = os.path.split(self.input_file_path)
            output_file_path = os.path.join(output_directory, output_file_path)
        else:
            output_file_path = self.input_file_path
        if stylesheet is not None:
            stylesheet = os.path.abspath(stylesheet)
            stylesheet = os.path.relpath(
                stylesheet,
                self.assets_directory,
                )
        self.report(
            output_file_path=output_file_path,
            stylesheet=stylesheet,
            verbose=verbose,
            )
        self._errored = False
        console = abjadbooktools.AbjadBookConsole(
            document_handler=self,
            locals=abjad.__dict__.copy(),
            )
        output_blocks = self.collect_output_blocks(
            self.input_file_contents,
            configuration=configuration,
            )
        if clean:
            input_blocks = {}
        else:
            input_blocks = self.collect_input_blocks(
                self.input_file_contents,
                stylesheet=stylesheet,
                )
            self.interpret_input_blocks(input_blocks, console)
        if not skip_rendering and not clean and not return_source:
            self.render_asset_output_proxies(
                input_blocks=input_blocks,
                )
        rebuilt_source = self.rebuild_source(
            configuration=configuration,
            input_blocks=input_blocks,
            output_blocks=output_blocks,
            input_file_contents=self.input_file_contents,
            )
        if return_source:
            return rebuilt_source
        self.write_rebuilt_source(
            rebuilt_source,
            output_file_path=output_file_path,
            )

    ### PUBLIC METHODS ###

    def collect_asset_output_proxies(self, input_blocks):
        from abjad.tools import abjadbooktools
        asset_output_proxies = []
        for code_block in input_blocks.values():
            for output_proxy in code_block.output_proxies:
                if isinstance(output_proxy, abjadbooktools.CodeOutputProxy):
                    continue
                asset_output_proxies.append(output_proxy)
        return asset_output_proxies

    def collect_input_blocks(self, input_file_contents, stylesheet=None):
        r"""Collects input blocks.

        ::

            >>> input_blocks = document_handler.collect_input_blocks(
            ...     input_file_contents.splitlines())

        ::

            >>> for source_line_range in input_blocks.keys():
            ...     print(source_line_range)
            ...
            (3, 5)
            (11, 15)
            (21, 23)

        ::

            >>> for input_block in input_blocks.values():
            ...     print(format(input_block))
            ...
            abjadbooktools.CodeBlock(
                ('print("hello, world!")',),
                starting_line_number=5,
                )
            abjadbooktools.CodeBlock(
                ("just_a_string = '''", 'show(Nothing!)', "'''"),
                starting_line_number=15,
                )
            abjadbooktools.CodeBlock(
                ('show(Note("c\'4"))',),
                starting_line_number=23,
                )

        """
        from abjad.tools import abjadbooktools
        input_blocks = collections.OrderedDict()
        in_input_block = False
        starting_line_number = None
        current_block_lines = None
        current_block_options = None
        for i, line in enumerate(input_file_contents):
            if line.startswith('<abjad'):
                if in_input_block:
                    message = 'Extra opening tag at line {}.'.format(i + 1)
                    raise ValueError(message)
                current_block_options = {'stylesheet': stylesheet}
                current_block_options.update(
                    self.extract_code_block_options(line),
                    )
                if line.startswith('<abjad>'):
                    in_input_block = True
                    current_block_lines = []
                    starting_line_number = i
                elif line.startswith('<abjadextract '):
                    starting_line_number = stopping_line_number = i
                    code_block = abjadbooktools.CodeBlock.from_latex_abjadextract_block(
                        line,
                        starting_line_number=i,
                        **current_block_options
                        )
                    source_line_range = (
                        starting_line_number,
                        stopping_line_number,
                        )
                    input_blocks[source_line_range] = code_block
            elif line.startswith('</abjad>'):
                if not in_input_block:
                    message = 'Extra closing tag at line {}'.format(i + 1)
                    raise ValueError(message)
                in_input_block = False
                stopping_line_number = i
                source_line_range = (
                    starting_line_number,
                    stopping_line_number,
                    )
                code_block = abjadbooktools.CodeBlock.from_latex_abjad_block(
                    current_block_lines,
                    starting_line_number=i,
                    **current_block_options
                    )
                input_blocks[source_line_range] = code_block
            elif in_input_block:
                current_block_lines.append(line)
        if in_input_block:
            raise ValueError('Unterminated tag at EOF.')
        return input_blocks

    def collect_output_blocks(self, input_file_contents, configuration=None):
        configuration = configuration or {}
        latex_configuration = configuration.get('latex', {})
        output_start_delimiter = latex_configuration.get(
            'output-start-delimiter',
            ('%%% ABJADBOOK START %%%',),
            )
        output_start_delimiter = output_start_delimiter[0]
        output_stop_delimiter = latex_configuration.get(
            'output-stop-delimiter',
            ('%%% ABJADBOOK END %%%',),
            )
        output_stop_delimiter = output_stop_delimiter[0]
        output_blocks = []
        in_output_block = False
        starting_line_number = None
        for i, line in enumerate(input_file_contents):
            if line.startswith(output_start_delimiter):
                if in_output_block:
                    message = 'Extra opening tag at line {}'.format(i + 1)
                    raise ValueError(message)
                in_output_block = True
                starting_line_number = i
            elif line.startswith(output_stop_delimiter):
                if not in_output_block:
                    message = 'Extra closing tag at line {}'.format(i + 1)
                    raise ValueError(message)
                in_output_block = False
                stopping_line_number = i
                source_line_range = (
                    starting_line_number,
                    stopping_line_number,
                    )
                output_blocks.append(source_line_range)
        return output_blocks

    @staticmethod
    def extract_code_block_options(source_line):
        r'''Extracts code block options.

        ::

            >>> source_line = '<abjad>'
            >>> document_handler.extract_code_block_options(source_line)
            {}

        ::

            >>> source_line = '<abjad>[hide=True]'
            >>> document_handler.extract_code_block_options(source_line)
            {'hide': True}

        ::

            >>> source_line = '<abjad>[strip_prompt=true, hide=false]'
            >>> options = document_handler.extract_code_block_options(
            ...     source_line)
            >>> for key, value in sorted(options.items()):
            ...     key, value
            ...
            ('hide', False)
            ('strip_prompt', True)

        ::

            >>> source_line = '<abjad>[allow_exceptions]'
            >>> options = document_handler.extract_code_block_options(
            ...     source_line)
            >>> for key, value in sorted(options.items()):
            ...     key, value
            ...
            ('allow_exceptions', True)

        '''
        options = {}
        line = source_line.strip()
        if '[' in line and line.endswith(']'):
            option_string = line.partition('[')[2][:-1]
            for part in option_string.split(','):
                part = part.strip()
                if '=' in part:
                    key, sep, value = part.partition('=')
                    key = key.lower().strip()
                    value = value.lower().strip()
                    if value == 'true':
                        options[key] = True
                    elif value == 'false':
                        options[key] = False
                    else:
                        options[key] = value
                else:
                    options[part] = True
        return options

    @classmethod
    def from_path(
        cls,
        input_file_path=None,
        assets_directory=None,
        latex_root_directory=None,
        ):
        assert os.path.exists(input_file_path)
        input_file_path = os.path.abspath(input_file_path)
        input_directory, _ = os.path.split(input_file_path)
        if not assets_directory:
            assets_directory = os.path.join(input_directory, 'assets')
        if not latex_root_directory:
            latex_root_directory = input_directory
        with open(input_file_path, 'r') as file_pointer:
            input_file_contents = file_pointer.read()
        document_handler = cls(
            input_file_contents=input_file_contents,
            input_file_path=input_file_path,
            assets_directory=assets_directory,
            latex_root_directory=latex_root_directory,
            )
        return document_handler

    def get_default_stylesheet(self):
        return None

    def interpret_input_blocks(self, input_blocks, console, verbose=True):
        code_blocks = tuple(input_blocks.values())
        if not code_blocks:
            return
        progress_indicator = systemtools.ProgressIndicator(
            message='    Interpreting code blocks',
            total=len(code_blocks),
            verbose=verbose,
            )
        with progress_indicator:
            for code_block in code_blocks:
                code_block.interpret(console)
                progress_indicator.advance()

    def rebuild_source(
        self,
        input_blocks,
        output_blocks,
        input_file_contents,
        configuration=None,
        ):
        input_file_contents = list(input_file_contents)
        blocks = list(input_blocks.items())
        output_blocks = ((_, None) for _ in output_blocks)
        blocks.extend(output_blocks)
        blocks.sort()
        for source_line_range, block in reversed(blocks):
            start, stop = source_line_range
            if block is None:
                input_file_contents[start:stop + 1] = []
                if start < len(input_file_contents) and \
                    not input_file_contents[start].strip():
                    input_file_contents[start:start + 1] = []
                continue
            output_lines = block.as_latex(
                configuration=configuration,
                output_directory=self.assets_directory,
                relative_output_directory=self.latex_assets_prefix,
                )
            if output_lines:
                output_lines.insert(0, '')
                stop += 1
                if stop < len(input_file_contents):
                    if '\\end{comment}' in input_file_contents[stop]:
                        stop += 1
                input_file_contents[stop:stop] = output_lines
        input_file_contents = '\n'.join(input_file_contents)
        input_file_contents = input_file_contents.splitlines()
        input_file_contents = (_.rstrip() for _ in input_file_contents)
        input_file_contents = '\n'.join(input_file_contents)
        return input_file_contents

    def register_error(self):
        self._errored = True

    def render_asset_output_proxies(
        self,
        input_blocks,
        ):
        asset_output_proxies = self.collect_asset_output_proxies(input_blocks)
        if not asset_output_proxies:
            return
        if not os.path.exists(self.assets_directory):
            os.makedirs(self.assets_directory)

        progress_indicator = systemtools.ProgressIndicator(
            message='    Writing assets',
            total=len(asset_output_proxies),
            )
        with progress_indicator:
            for asset_output_proxy in asset_output_proxies:
                asset_output_proxy.render_for_latex(
                    self.assets_directory,
                    )
                progress_indicator.advance()

    def report(
        self,
        output_file_path=None,
        stylesheet=None,
        verbose=None,
        ):
        if verbose:
            print('Processing {}:'.format(self.input_file_path))
            print('    output file:     {}'.format(output_file_path))
            print('    asset directory: {}'.format(self.assets_directory))
            print('    asset prefix:    {}'.format(self.latex_assets_prefix))
            print('    latex root:      {}'.format(self.latex_root_directory))
            if stylesheet is not None:
                print('    stylesheet:      {} ({})'.format(
                    stylesheet,
                    os.path.abspath(os.path.join(
                        self.assets_directory,
                        stylesheet,),)
                    ))
            else:
                print('    stylesheet:      None')
        else:
            print('Processing {}'.format(self.input_file_path))

    def unregister_error(self):
        self._errored = False

    def write_rebuilt_source(
        self,
        rebuilt_source,
        output_file_path=None,
        ):
        if os.path.exists(output_file_path):
            with open(output_file_path, 'r') as file_pointer:
                old_source = file_pointer.read()
            if old_source == rebuilt_source:
                return
        with open(output_file_path, 'w') as file_pointer:
            file_pointer.write(rebuilt_source)

    ### PUBLIC PROPERTIES ###

    @property
    def assets_directory(self):
        return self._assets_directory

    @property
    def console(self):
        return self._console

    @property
    def errored(self):
        return self._errored

    @property
    def input_directory(self):
        if self.input_file_path is None:
            return '.'
        return os.path.dirname(self.input_file_path)

    @property
    def input_file_contents(self):
        return self._input_file_contents

    @property
    def input_file_path(self):
        return self._input_file_path

    @property
    def latex_assets_prefix(self):
        assets_directory = self.assets_directory
        if not assets_directory:
            assets_directory = 'assets'
        if self.latex_root_directory:
            return os.path.relpath(
                assets_directory,
                self.latex_root_directory,
                )
        return os.path.relpath(
            assets_directory,
            self.input_directory,
            )
    @property
    def latex_root_directory(self):
        return self._latex_root_directory




# -*- coding: utf-8 -*-
import copy
from abjad.tools import scoretools
from abjad.tools import sequencetools
from abjad.tools.quantizationtools.QTarget import QTarget
from abjad.tools.topleveltools import attach
from abjad.tools.topleveltools import select


class BeatwiseQTarget(QTarget):
    r'''A beat-wise quantization target.

    Not composer-safe.

    Used internally by ``Quantizer``.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    ### PRIVATE METHODS ###

    def _notate(
        self,
        attach_tempos=True,
        attack_point_optimizer=None,
        grace_handler=None,
        ):
        voice = scoretools.Voice()

        # generate the first
        beat = self.items[0]
        components = beat.q_grid(beat.beatspan)
        if attach_tempos:
            attachment_target = components[0]
            leaves = select(attachment_target).by_leaf()
            if isinstance(attachment_target, scoretools.Container):
                attachment_target = leaves[0]
            tempo = copy.copy(beat.tempo)
            attach(tempo, attachment_target)
        voice.extend(components)

        # generate the rest pairwise, comparing tempi
        for beat_one, beat_two in sequencetools.Sequence(self.items).nwise():
            components = beat_two.q_grid(beat_two.beatspan)
            if (beat_two.tempo != beat_one.tempo) and attach_tempos:
                attachment_target = components[0]
                leaves = select(attachment_target).by_leaf()
                if isinstance(attachment_target, scoretools.Container):
                    attachment_target = leaves[0]
                tempo = copy.copy(beat_two.tempo)
                attach(tempo, attachment_target)
            voice.extend(components)

        # apply logical ties, pitches, grace containers
        self._notate_leaves(
            grace_handler=grace_handler,
            voice=voice,
            )

        # partition logical ties in voice
        attack_point_optimizer(voice)

        return voice

    ### PUBLIC PROPERTIES ###

    @property
    def beats(self):
        r'''Beats of beatwise q-target.
        '''
        return tuple(self.items)

    @property
    def item_class(self):
        r'''Item class of beatwise q-target.
        '''
        from abjad.tools import quantizationtools
        return quantizationtools.QTargetBeat



# -*- coding: utf-8 -*-
import platform
import pytest
import sys
from abjad import *


#@pytest.mark.skipif(
#    platform.python_implementation() != 'CPython',
#    reason='Benchmarking is only for CPython.',
#    )
#def test_systemtools_IOManager_count_function_calls_01():
#    result = systemtools.IOManager.count_function_calls(
#        "Note('c4')",
#        globals(),
#        )
#    if sys.version_info[0] == 2:
#        assert result < 14000
#    else:
#        assert result < 25000


#@pytest.mark.skipif(
#    platform.python_implementation() != 'CPython',
#    reason='Benchmarking is only for CPython.',
#    )
#def test_systemtools_IOManager_count_function_calls_02():
#    result = systemtools.IOManager.count_function_calls(
#        "Note(-12, (1, 4))",
#        globals(),
#        )
#    if sys.version_info[0] == 2:
#        assert result == 188
#    else:
#        # inequality because Pythons 3.2, 3.3, 3.4, 3.5 vary somewhat
#        assert result <= 205
